; =============================================================================
;  * Programm "Multifunktionsschalter fr Rhren(vor-)verstrker"            *
; MCU:		Atmel AT90S2313
; Takt:		4,0 MHz
; Display:	STN 1x16 LCD
;
; Version:	0.9b (zuletzt bearbeitet am 03.02.2004)
; =============================================================================
;
;  * Belegung der Register * 
;  r0 bis r15:	Temporre Nutzung
;  r16, r17:	Temporre Nutzung
;  r18:		Zeichen fr Textausgabe
;  r19:		BIT 0,1,2,3: Gewhlter Kanal
;			BIT 4: Verstrker eingeschaltet
;			BIT 5: Kanal "CD-DIREKT" mit Sonderfunktionen
;  r20, r21, r22, r23:
;			32Bit-Zhlerstand Betriebszeit in Sekunden
;  r24,r25:	bergabe einer 16Bit Zahl z.B. fr Display-Darstellung
;  r29:		Zhlt auf LCD ausg. Zeichen, da dieses virtuell 2x8 Zeilen anzeigt
;  r26,r27:	16Bit-Nutzung als Pointer:
;  r30 r31:	-- " -- Zugriff auf Programmspeicher = Z
;  Z		bergabe einer Textadresse oder einer Zahl fr Display-Ausgabe
;
;  * Belegung der Portbits * BIT / RICHTUNG / FUNKTION
;  PB0	OUT	Netzteil Verstrker
;  PB1	IN	Wahlschalter OFF-Stellung - Betriebszeit im EEPROM ablegen
;  PB2	IN	Taste Rhrenwechsel (Betriebszeit-Reset, Druck 3s im CD-DIREKT Modus)
;  PB3	OUT	DUO-LED, Kanal Rot
;  PB4	OUT	DUO-LED, Kanal Grn
;  PD0	IN	Wahlschalter Pos. CD
;  PD1	IN	Wahlschalter Pos. PHONO
;  PD2	IN	Wahlschalter Pos. TAPE
;  PD3	IN	Wahlschalter Pos. TUNER
;  PD4	OUT	Schieberegister (fr Display) CLOCK
;  PD5	OUT	Schieberegister BIT sowie Display REGISTER SELECT
;  PD6	OUT	Display ENABLE
;
; *****************************************************************************
;   Funktion:
;   - Nach dem Einschalten wird die Modellnummer angezeigt
;   - Wird das Netzteil-Relais eingeschaltet, "Heizen" anzeigen. 
;     Verstrker wird erst wieder in Stellung "STANDBY" abgeschaltet,
;     NICHT in Stellung CD-DIREKT - um ihn beim schnellen Schalten zwischen
;     den Betriebspositionen zu schonen und eine zwischenzeitliche Anzeige
;	  der Betriebszeit zu ermglichen
;   - Anzeige des gewhlten Kanals,
;	  in Stellung CD-DIREKT auch Anzeige der Rhrenbetriebszeit in Stunden
;   - Zhlen der Rhrenbetriebszeit in Sekunden
; *****************************************************************************

; LCD Steuerung
.equ LCD_E		= 6
.equ LCD_RS		= 5
.equ LCD_DATA	= 5
.equ LCD_CLK	= 4

; Sonstige Portbits
.equ ONOFF  	= 0
.equ STANDBY	= 1
.equ EERESET	= 2
.equ RED		= 3
.equ GREEN		= 4
.equ K_CD		= 0
.equ K_PHONO	= 1
.equ K_TAPE		= 2
.equ K_TUNER	= 3

; werte fr Timer1
.equ timercmp	= 15625	;Entspricht 1 Sek. bei Vorteiler 256 und 4MHz

.include "2313def.inc"	;Definitionsdatei laden
.cseg					;Beginn eines Code-Segmentes
.org 0					;...ab Adresse 0
rjmp	start			;Tabellen etc. berspringen

; === Interrupt-Vektoren ======================================================
exint0:	reti
exint1:	reti
t1icap:	reti
t1comp:	rjmp addSek			;Timer1, jede Sekunde
t1ovfl:	reti
t0ovfl:	reti
uarrxc:	reti
uartxe:	reti
uartxc:	reti
ancomp:	reti

; === Texte fr die Displayausgabe, Tabellen ==================================
intro_a:		.db	"Schawaller VV-01",0
t_std:			.db	"h",0
t_warm:			.db	"Heizen: ",0
t_cddir:		.db	" CD-DIR ",0
t_cd:			.db	"       CD",0
t_phono:		.db	"     PHONO",0
t_tuner:		.db	"     TUNER",0
t_tape:			.db	"      TAPE",0
t_reset:		.db	" Speicher RESET",0
t_nullh:		.db	"Betriebszeit: 0h",0
; =============================================================================

start:	cli					;Alle Interrupts - damit nat. auch den Timer - deaktivieren

		ldi	r16,low(ramend)	;Stackpointer...
		out	spl,r16			;...auf RAM-Ende setzen
		ldi	r16,0b00011001	;Port B: I/0...
		out	ddrb,r16		;...konfigurieren
		ldi	r16,0b01110000	;Port D: I/0...
		out	ddrd,r16		;...konfigurieren

		ldi	r16,(1<<ctc1)+(1<<cs12)	;Timer1 Reset bei Interrupt,
		out	tccr1b,r16				;Vorteiler = 256
		ldi	r16,low(timercmp)	;Timer1-Compare fr 1 Sek
		ldi	r17,high(timercmp)
		out	ocr1ah,r17			;ausgeben
		out	ocr1al,r16
		ldi	r16,(1<<ocie1a)		;Timer1 Compare-Interrupt
		out	timsk,r16			;aktivieren

		cbi PORTB,ONOFF		;Um sicher zu gehen: Verstrker ausschalten
		clr	r19				;= Verstrker AUS, gew. Kanal KEINER

; ********************** Endlosschleife / Hauptprogramm ***********************
loop:   sbis PINB,STANDBY	;ist Schalter in Stellung "AUS"?
		rjmp vv_an			;nein -> betriebsbereit machen
istaus:	sbrc r19,4			;ja -> Ist das bereits bekannt?
		rcall p_off			;nein -> Abschalten, Zeit loggen etc.
		rjmp loop			;erneut prfen
vv_an:	rcall init			;LCD einschalten, Arbeitszeit lesen etc.
loop2:  sbis PIND,K_CD		;Kanal CD?
		rcall c_cd
		sbis PIND,K_PHONO	;Kanal PHONO?
		rcall c_phono
		sbis PIND,K_TAPE	;Kanal TAPE?
		rcall c_tape
		sbis PIND,K_TUNER	;Kanal TUNER?
		rcall c_tuner
		sbis PIND,K_CD		;War Kanal (CD) gesetzt?
		rjmp k_ges
		sbis PIND,K_PHONO	;s.o.
		rjmp k_ges
		sbis PIND,K_TAPE	;s.o.
		rjmp k_ges
		sbis PIND,K_TUNER	;s.o.
		rjmp k_ges
		rcall c_cddir		;Hier angekommen = Kein Kanal gewhlt = CD-Direkt
k_ges:	sbis PINB,STANDBY	;ist Schalter in Stellung "AUS"?
		rjmp loop2			;nein -> weitermachen...
		cbi	PORTB,RED		;LED "Rot" abschalten, falls Verstrker aus / CD-DIR
		rjmp istaus			;In den "Aus-Modus" schicken
; ********************** ENDE Endlosschleife **********************************

c_cd:		sbrc r19,0			;Bereits bekannt, dass Kanal CD gesetzt?
			ret					;ja -> nichts zu tun
			rcall p_on			;Verstrker ggf. einschalten
			ldi r19, 0b00010001 ;Kanal-Bit in Statusregister setzen
			ldi	zl,low(2*t_cd)  ;Kanal-Text addressieren...
			ldi	zh,high(2*t_cd)
			rcall lcd_clear		;...anzeigen
			rcall lcd_print
			ret

c_phono:	sbrc r19,1			;Bereits bekannt, dass Kanal PHONO gesetzt?
			ret					;ja -> nichts zu tun
			rcall p_on			;Verstrker ggf. einschalten
			ldi r19, 0b00010010 ;Kanal-Bit in Statusregister setzen			
			ldi	zl,low(2*t_phono)  ;Kanal-Text addressieren...
			ldi	zh,high(2*t_phono)
			rcall lcd_clear		;...anzeigen
			rcall lcd_print
			ret

c_tape:		sbrc r19,2			;Bereits bekannt, dass Kanal TAPE gesetzt?
			ret					;ja -> nichts zu tun
			rcall p_on			;Verstrker ggf. einschalten
			ldi r19, 0b00010100 ;Kanal-Bit in Statusregister setzen
			ldi	zl,low(2*t_tape)  ;Kanal-Text addressieren...
			ldi	zh,high(2*t_tape)
			rcall lcd_clear		;...anzeigen
			rcall lcd_print
			ret

c_tuner:	sbrc r19,3			;Bereits bekannt, dass Kanal TUNER gesetzt?
			ret					;ja -> nichts zu tun
			rcall p_on			;Verstrker ggf. einschalten
			ldi r19, 0b00011000 ;Kanal-Bit in Statusregister setzen
			ldi	zl,low(2*t_tuner)  ;Kanal-Text addressieren...
			ldi	zh,high(2*t_tuner)
			rcall lcd_clear		;...anzeigen
			rcall lcd_print
			ret

c_cddir:	sbrc r19,5			;Bereits bekannt, dass Kanal CD-DIREKT gesetzt?
			rjmp cdd_ende		;ja -> Kanal wird bereits angezeigt
			andi r19,0b11110000   ;Fremdkanal-Bits lschen...
			ori r19, 0b00100000   ;...Kanal-Bit setzen
			ldi	zl,low(2*t_cddir) ;Kanal-Text addressieren...
			ldi	zh,high(2*t_cddir)
			rcall lcd_clear		;Zeigt Kanal (Text ab (Z)) und Betriebszeit an
			rcall lcd_print
			rcall calc_betr		;Betriebszeit in Stunden umrechnen (Ergebnis in r25;r24)
			rcall print_dec_lcd	;Betriebszeit ausgeben...
			ldi	zl,low(2*t_std)	;...gefolgt...
			ldi	zh,high(2*t_std);...vom Zeichen...
			rcall lcd_print		;...h fr Stunden
cdd_ende:	sbic PINB,EERESET	;Taste fr EEPROM-Reset / Rhrenwechsel gedrckt?
			ret					;nein -> nichts mehr zu tun
			rcall lcd_clear		;LCD lschen
			ldi	zl,low(2*t_reset)  ;Text "Speicher-Reset" addressieren...
			ldi	zh,high(2*t_reset)
			rcall lcd_print		;...und ausgeben.
			rcall wait3			;3 Sekunden warten
			sbic PINB,EERESET	;Taste immer noch gedrckt?
			rjmp cdd_abbr		;nein -> war wohl ein Versehen...
			cli					;ja -> Interrupts stoppen...
			clr r20				;...Zhler lschen...
			clr r21
			clr r22
			clr r23
			rcall wEepr			;und in EEPROM schreiben
			sbrc r19,4			;Verstrker eingeschaltet?
			sei					;ja -> Interrupts/Timer wieder aktivieren
			ldi	zl,low(2*t_nullh)  ;Text "Betriebszeit: 0h"...
			ldi	zh,high(2*t_nullh) ;...addressieren...
			rcall lcd_clear
			rcall lcd_print
			rcall wait3			;...und fr 3 Sekunden ausgeben
cdd_abbr:	ldi	zl,low(2*t_cddir)  ;Kanal-Text addressieren...
			ldi	zh,high(2*t_cddir)
			rcall lcd_clear		;Anzeige restaurieren
			rcall lcd_print
			rcall calc_betr
			rcall print_dec_lcd
			ldi	zl,low(2*t_std)
			ldi	zh,high(2*t_std)
			rcall lcd_print
			ret					;ENDE C_CDDIR

init:	clr	r19				;Verstrker ist AUS, gew. Kanal KEINER
		sbi	PORTB,RED     	;rote LED-Farbe einschalten
		rcall rEepr			;Betriebszeit aus dem Eeprom lesen
		rcall lcd_init		;LCD initialisieren
		rcall intro			;Begruessungstext anzeigen
		ret

p_on:	sbrc r19,4			;Noch ausgeschaltet?
		ret					;nein -> nichts zu tun
		sbi PORTB,ONOFF		;ja -> einschalten
		sei					;Interrupts einschalten => Timer1 aktiv
		rcall warmup		;"Heizen"-Animation einblenden
		ori r19, 0b00010000	;Statusbit "EIN" setzten
		ret					;ENDE P_ON

p_off:	clr	r19				;= Verstrker AUS, gew. Kanal KEINER
		cbi PORTB,ONOFF		;ausschalten
		cbi	PORTB,RED		;LEDs abschalten
		cbi	PORTB,GREEN
		rcall wEepr			;Betriebszeit in den EEPROM schreiben
		rcall wait05
		sbi	PORTB,GREEN		;Blinken, um erfolgreiches Schreiben
		rcall wait05		;des EEPROM anzuzeigen. Nach diversen
		cbi	PORTB,GREEN		;rtselhaften Abstrzen eingebaut -
		rcall wait05		;sollte im aktuellen Schaltungsdesign
		sbi	PORTB,GREEN		;aber nicht mehr dazu kommen.
		rcall wait05
		cbi	PORTB,GREEN
		rcall wait3			;gegen Spannungsspitzen beim Wiedereinschalten
		ret

;Warteschleife (grob 0,1s bei 4.0MHz)
w100ms:  clr	r1			;High-Zhler setzen
		 clr	r2			;Low-Zhler setzen
w100ms1: dec	r2			;Low-Zhler-1
		 nop
		 nop
		 brne	w100ms1		;Schleife
		 dec	r1			;High-Zhler-1		 
		 brne	w100ms1		;Schleife
		 ret				;ENDE W100MS
;Warteschleife (ungefhr 1/2s bei 4.0MHz)
wait05:	ldi	r16,10			;r18, r17 und r16
		clr	r15				;ergeben zusammen
		clr	r14				;einen 3-Byte-Zhler
wait1:	dec	r14				;niedrigstes Byte -1
		brne wait1			;0 erreicht? nein -> Schleife
		dec	r15				;mittleres Byte -1
		brne wait1			;0 erreicht? nein -> Schleife
		dec	r16				;hchstes Byte -1
		brne wait1			;0 erreicht? nein -> Schleife
		ret					;ENDE WAIT1
wait15:	rcall wait05		;1,5s warten
		rcall wait05
		rcall wait05
		ret					;ENDE WAIT15
wait3:	rcall wait15		;3s warten
		rcall wait15
		ret					;ENDE WAIT3

intro:	rcall lcd_clear			;LCD lschen
		ldi	zl,low(2*intro_a)	;Zeiger auf Text
		ldi	zh,high(2*intro_a)	;...
		rcall lcd_print			;...ausgeben
		rcall wait15			;1,5s warten
		ret						;ENDE INTRO

warmup: rcall lcd_clear			;LCD lschen
		sbi	PORTB,GREEN			;grne LED-Farbe zuschalten = gelb
		ldi	zl,low(2*t_warm)	;Zeiger auf Text
		ldi	zh,high(2*t_warm)	;...
		rcall lcd_print			;...ausgeben
		ldi r18, 0xff			;Zeichen "Balken"
		rcall lcd_w_data		;...ausgeben...
		rcall wait05
		ldi r18, 0xff			;...und zwar 8 mal, mit einem Zeitabstand von
		rcall lcd_w_data		;je 1/2s, so dass sich eine "Animation"
		rcall wait05			;von 4 Sekunden Dauer ergibt
		ldi r18, 0xff
		rcall lcd_w_data
		rcall wait05
		ldi r18, 0xff
		rcall lcd_w_data
		rcall wait05
		ldi r18, 0xff
		rcall lcd_w_data
		rcall wait05
		ldi r18, 0xff
		rcall lcd_w_data
		rcall wait05
		ldi r18, 0xff
		rcall lcd_w_data
		rcall wait05
		ldi r18, 0xff
		rcall lcd_w_data
		rcall wait05
		cbi	PORTB,RED			;rote LED-Farbe ausschalten = grn
		ret

; === LCD Routinen ============================================================

;Schiebt r18 ins 74xx164. Inhalt von r18 wird dabei zerstrt
lcd_ld:		ldi r16, 0x07		;Zhler = "8 Bits"
lcd_ld_1:	cbi PORTD,LCD_DATA	;LCD_DATA erstmal lschen,
			sbrc r18,0x07		;r18 MSB gesetzt?
			sbi PORTD,LCD_DATA	;Ja -> LCD_DATA = 1
			nop
  			sbi PORTD,LCD_CLK	;Takt auf LCD_CLK
  			lsl r18				;Whrend der Wartezeit r18 shiften
  			nop					;Viel zu lange warten - so
  			nop					;funktioniert es auch mit
  			nop					;antiquarischen ICs
  			nop
  			cbi  PORTD,LCD_CLK  ;Takt-Ende
			nop
			nop
  			nop
			cpi r16,0x00		;Schon bei Bit 0?
			breq lcd_ld_2		;Ja -> Ende
			dec r16				;Nein -> dec(Zhler)
			rjmp lcd_ld_1		;weitermachen
lcd_ld_2:	ret

lcd_enable:	sbi PORTD,LCD_E		;LCD-Datenbernahme
			nop					;"Vorsichtshalber" warten...
			nop
			nop
			nop
			nop
			nop
			nop
  			cbi PORTD,LCD_E		;...und Bit rcksetzen
			clr r0				;Warten auf LCD nach Befehl oder Zeichen
lcd_wait:	nop
  			nop
  			nop
  			nop
  			nop
  			nop
  			nop
  			nop
  			nop
  			nop
  			dec r0
			brne lcd_wait
  			ret

lcd_w_cmd:	rcall lcd_ld		;Wert in Register laden...
			cbi PORTD,LCD_RS	;...LCD-Kommando schreiben...
			rcall lcd_enable	;...und Datenbernahme veranlassen
			ret

lcd_w_data: rcall lcd_ld		;Wert in Register laden...
			sbi PORTD,LCD_RS	;LCD-Daten schreiben
			rcall lcd_enable	;...und Datenbernahme veranlassen
			ret

lcd_clear:	ldi r18,0b00000001	;Display lschen, DD RAM - Counter rcksetzen
  			rcall lcd_w_cmd
			ldi r18,0b00000010  ;Cursor-home, ggf. Shift restaurieren
			rcall lcd_w_cmd
			clr r29				;8-Zeichen-Zhler auf 0
  			ret

lcd_init:	rcall w100ms		;Nach "Strom an" dauert es, bis LCD bereit ist.
			cbi PORTD,LCD_RS
  			cbi PORTD,LCD_CLK
  			cbi PORTD,LCD_E
  			ldi r18,0x38		;8-Bit, 2 Zeilen (virt. 2x8, "Billigdisp.") , 5x8 Pixel
  			rcall lcd_w_cmd
  			ldi r18,0x0c		;Cursorblinken AUS, Cursor AUS, Display AN
  			rcall lcd_w_cmd
  			ldi r18,0x06		;Shift aus, Increment
  			rcall lcd_w_cmd
			rcall lcd_clear		;Adressreg. auf DD-RAM Addr. 0 setzen
  			ret

;Gibt einen String ab Adresse Z aus, Schlussbyte = 0
lcd_print:	lpm					;Zeichen aus dem Programmspeicher laden...
			mov r18,r0			;...und in Reg. r18 kopieren
			cpi r18,0x00		;r18 = 0 = Textende?
			breq lcd_print2		;ja -> fertig
			rcall lcd_w_data	;nein -> ausgeben...
			inc r29				;Zhler fr Zeichenzahl + 1
			cpi r29,0x08		;War dies das 8. Zeichen?
			brne lcd_print1		;nein -> weiter machen
			ldi r18,0xC0		;ja -> virtuellen Zeilensprung...
			rcall lcd_w_cmd		;...veranlassen
lcd_print1:	adiw zl,1			;Adresse += 1
			rjmp lcd_print		;...und nchstes Zeichen laden
lcd_print2:	ret

;Gibt 16-Bit-Wert in r25:r24 als linksbndige Dezimalzahl aus. Wert wird zerstrt.
;Zugegeben faul implementiert, aber es ist immer noch viel Platz im Flash-ROM...
print_dec_lcd:  clr r1			;r1 soll fhrende Nullen aufdecken
				ldi r18,0x30	;Dezimalstelle ASCII-Offset laden (48="0")
print_l_1:		cpi	r25,0x27	;MSB mit MSB von 10 000 vergleichen
				brlo print_l_1p	;ist kleiner -> nichts (mehr) zu tun
				breq print_l_1e	;ist gleich -> Wert knnte auch < 10000 sein
				rjmp print_l_1g
print_l_1e:		cpi r24, 0x10	;ist im gleich-Fall das LSB "gro genug"?
				brlo print_l_1p	;nein -> fertig
print_l_1g:		inc r18			;war insg. grer -> passt mindestens 1 mal
				subi r25, 0x27	;Zuerst MSB von 10 000 abziehen
				cpi r24, 0x10	;LSB kleiner?
				brlo print_l_1k	;ja -> Komplement addieren
				subi r24,0x10	;nein -> einfach abziehen...
				rjmp print_l_1x	;...Subtraktion abgeschlossen
print_l_1k:		dec r25			;bertrag von MSB abziehen
				ldi r26,0xf0	;Hilfsregister mit FF + 1 - Dez.LSB initialisieren...
				add r24,r26		;...aufaddieren, fertig.
print_l_1x:		rjmp print_l_1	;Schleife
print_l_1p:		sbrc r1, 0		;Bit 0 in r1 gesetzt?
				rjmp print_l_1a	;ja -> handelt sich nicht um fhrende 0
				cpi r18, 0x30	;r18 = fhrende "0"?
				breq print_l_1f	;ja -> nicht anzeigen
				inc r1			;nein -> Erste Ziffer != 0 entdeckt
print_l_1a:		rcall lcd_w_data ;-> anzeigen
print_l_1f:		ldi r18,0x30
print_l_2:		cpi	r25,0x03	;s.o., jedoch "1000er"
				brlo print_l_2p
				breq print_l_2e
				rjmp print_l_2g
print_l_2e:		cpi r24, 0xE8
				brlo print_l_2p
print_l_2g:		inc r18
				subi r25, 0x03
				cpi r24, 0xE8
				brlo print_l_2k
				subi r24,0xE8
				rjmp print_l_2x
print_l_2k:		dec r25
				ldi r26,0x18	;FF + 1 - Dez. LSB (E8)
				add r24,r26
print_l_2x:		rjmp print_l_2
print_l_2p:		sbrc r1, 0
				rjmp print_l_2a
				cpi r18, 0x30
				breq print_l_2f
				inc r1
print_l_2a:		rcall lcd_w_data
print_l_2f:		ldi r18,0x30	;s.o., jedoch "100er", etwas anders:
print_l_3:		cpi	r24,0x64	;LSB > 100?
				brlo print_l_3e	;nein -> ...
				rjmp print_l_3g	;ja -> passt >= 1 mal
print_l_3e:		cpi r25, 0x00	;... -> "noch MSB-Reste brig?"
				breq print_l_3p ;auch nicht -> fertig
print_l_3g:		inc r18			;ab hier analog
				cpi r24, 0x64
				brlo print_l_3k
				subi r24,0x64
				rjmp print_l_3x
print_l_3k:		dec r25
				ldi r26,0x9c	;FF + 1 - Dez. LSB (64)
				add r24,r26
print_l_3x:		rjmp print_l_3
print_l_3p:		sbrc r1, 0
				rjmp print_l_3a
				cpi r18, 0x30
				breq print_l_3f
				inc r1
print_l_3a:		rcall lcd_w_data
print_l_3f:		ldi r18,0x30	;s.o., jedoch "10er". MSB interessiert nicht mehr
print_l_4:		cpi	r24,0x0A	;LSB > 10?
				brlo print_l_4p	;nein -> fertig
print_l_4g:		inc r18			;ja -> passt
				subi r24,0x0A
print_l_4x:		rjmp print_l_4
print_l_4p:		sbrc r1, 0
				rjmp print_l_4a
				cpi r18, 0x30
				breq print_l_4f
				inc r1
print_l_4a:		rcall lcd_w_data
print_l_4f:		ldi r18,0x30	;Letzte - 5te - Dezimalstelle, besonders einfach
print_l_5:		cpi	r24,0x01	;LSB > 1?
				brlo print_l_5p	;nein -> fertig
print_l_5g:		inc r18			;ja -> passt
				dec r24
print_l_5x:		rjmp print_l_5
print_l_5p:		rcall lcd_w_data  ;Wird auf jeden Fall ausgegeben
print_l_x:		ret
; === LCD Routinen ENDE =======================================================

;Rechnet die Betriebszeit in Stunden um (Erg. in r25, r24)
;Aufbau Reg.Sekunden:	r23;r22;r21;r20
;Aufbau Kopie:  		r27;r26;r25;r24
;Divisor (3600) 	  = 0E;10
;Hilsregister r16 (Zhler) und r17 fr 8-Bit Addition
calc_betr:
		brie calci;			;Interrupts (Timer1) aktiv?
		mov r27,r23			;nein -> 32-Bit-Zhler kann
		mov r26,r22			;gefahrlos kopiert werden.
		mov r25,r21
		mov r24,r20
		rjmp calc2			;weiter
calci:	cli					;SONST: Interrupts deaktivieren
		mov r27,r23			;Arbeit
		mov r26,r22
		mov r25,r21
		mov r24,r20
		sei					;Interrupts wieder starten
calc2:	ldi r16,0x10		;16 Schritte
loop_d:	cpi r27,0x0E		;Vergleich r27 mit MSB Divisor
		brlo wt_kl			;Kleiner / Divisor "passt" nicht
		cpi r26, 0x10		;"Byteweise Subtraktion mglich?"
		brlo sub_kp			;r26 > 256 => komplizierter (kp soll aber fr "Komplement" stehen)
		subi r26, 0x10		;sonst einfach -
		subi r27, 0x0E		;byteweise abziehen...
		rjmp wt_gr			;...fertig.
sub_kp:	ldi r17, 0xF0		;LSB X = r26 = 255 - 16 + LSB + 1
		add r26, r17		;aufaddieren.
		subi r27, 0x0E		;MSB -= 14
		dec r27				;bertrag abziehen
wt_gr:	rcall shiften		;32Bit-Linksshift
		ori r24, 0x01		;Divident war grer => Divisor passt mind. 1mal => 1 ablegen...
		rjmp wetr_cb		;...weitermachen
wt_kl:	rcall shiften		;
        andi r24, 0xFE		;Divident war kleiner => "Divisor passt 0mal" 0 =>
wetr_cb: cpi r16, 0		    ;Schon 16 Schritte?
		breq ende_d			;ja -> Fertig, Ergebnis bereits in r25;r24
		dec r16
		rjmp loop_d			;nein -> weitermachen
ende_d:	ret					;ENDE CALC_BETR

;Hilfs-Sub fr CALC_BETR, linksshift r25;r24;r2;r1
shiften: lsl r27			;kann "gefahrlos" geschiftet werden / Bit 7 interessiert nicht
        sbrc r26,7			;r26 MSB gesetzt?
		ori r27,0x01		;ja -> bertragen
		lsl r26				;u.s.w.
        sbrc r25,7
		ori r26,0x01
		lsl r25
        sbrc r24,7
		ori r25,0x01
		lsl r24
		ret					;ENDE SHIFTEN

; INTERRUPT; Addiert eine Sekunde auf den Zeitspeicher
addSek:	cpi	r20, 0xff		;Byte 1 voll?
		breq add_u1			;Ja - bertrag
		inc r20				;Nein - erhhen, fertig
		reti
add_u1:	clr r20				;Byte 1 lschen...
		cpi	r21, 0xff		;u.s.w.
		breq add_u2
		inc r21
		reti
add_u2:	clr r21
		cpi r22, 0xff
		breq add_u3
		inc r22
		reti
add_u3:	clr r22
		inc r23				;>136 Jahre => berlauf unwahrscheinlich :-)
		reti				;ENDE ADDSEK

; Liest die Rhrenbetriebszeit aus dem EEPROM, stoppt ggf. die Interrupts
rEepr:	cli					;Interrupts stoppen
		ldi r26, zeit_betr	;EEPROM Addresse auf erstes Byte
		rcall ee_rd			;...lesen...
		mov r20, r16		;...kopieren
		adiw r26,1			;nchstes Byte
		rcall ee_rd			;s.o.
		mov r21, r16
		adiw r26,1
		rcall ee_rd
		mov r22, r16
		adiw r26,1
		rcall ee_rd
		mov r23, r16
		ret					;ENDE REEPR

; Schreibt die Rhrenbetriebszeit in den EEPROM, stoppt ggf. die Interrupts
wEepr:	cli					;Interrupts stoppen
		ldi r26, zeit_betr	;EEPROM Addresse auf erstes Byte
		mov r16, r20		;...kopieren
		rcall ee_wr			;...schreiben...
		adiw r26,1			;nchstes Byte
		mov r16, r21
		rcall ee_wr			;s.o.
		adiw r26,1
		mov r16, r22
		rcall ee_wr
		adiw r26,1
		mov r16, r23
		rcall ee_wr
		ret					;ENDE WEEPR

;EEPROM lesen - r26: Adresse, r16: EEPROM-Daten, r1 wird gelscht
ee_rd:	sbic eecr,eewe	;Schreibzugriff?
		rjmp ee_rd		;ja -> warten
		out eear,r26	;Adresse bergeben
		sbi eecr,eere	;Lese-Bit setzen
		in r16,eedr		;Byte auslesen
		clr r1
		out eear,r1		;EEPROM Adresse zurcksetzen
		ret

;EEPROM schreiben - r26: Adresse, r16: EEPROM-Daten, r1 wird gelscht
ee_wr:	sbic eecr,eewe	;Schreibzugriff?
		rjmp ee_wr		;ja -> warten
		out eear,r26	;Adresse bergeben
		out eedr,r16	;Datenbyte setzen
		sbi eecr,eemwe	;Masterwrite-Bit setzen
		sbi eecr,eewe	;Write-Bit setzen
ee_wr2:	sbic eecr,eewe	;fertig?
		rjmp ee_wr2	    ;nein -> Schleife
		clr	r1
		out	eear,r1		;EEPROM Adresse zurcksetzen
		ret

; === EEPROM-Bereich ==========================================================
.eseg
dummy:		.db	0xff,0xff,0xff,0xff,0xff,0xff
			;...ich habe mir diverse Assemblerprogramme mit EEPROM-Nutzung
			;angesehen - berall waren dummy-Bytes am Anfang.
			;Vielleicht um Abnutzung whrend der Testphase auszugleichen?
			;Auf jeden Fall ist hier noch mehr als genug Platz brig, ein paar
			;unntze Bytes knnen zumindest nicht schaden...
zeit_betr:	.db	0x00,0x00,0x00,0x00	;32Bit-Zhler fr Rhren-Betriebszeit in Sekunden
; =============================================================================